import os
import ssl
from collections import Counter
from datetime import datetime
from typing import List, Any, Dict, Iterator
import dotenv
from elasticsearch_dsl import connections, Document, Text, Integer, Search, Date
from lxml import etree

speaker_index: str = "redner"
protocol_index: str = "dbtplenarprotokoll"
protocols_folder_path: str = "/home/laudatio/dbt_protocols_period19"
# load environment variables
dotenv_file: str = "/home/laudatio/.env"
dotenv.load_dotenv(dotenv_file)
user: str = os.environ["ELASTICSEARCH_USER"]
pw: str = os.environ["ELASTICSEARCH_PASS"]
scheme: str = os.environ["ELASTICSEARCH_SCHEME"]
host: str = os.environ["ELASTICSEARCH_HOST"]
port: str = os.environ["ELASTICSEARCH_PORT"]
certs_dir: str = os.environ["CERTS_DIR"]
# get paths to certificates
server_cert_path: str = os.path.join(certs_dir, "ca", "ca.crt")
kibana_certs_path: str = os.path.join(certs_dir, "kibana")
client_cert_path: str = os.path.join(kibana_certs_path, "kibana.crt")
client_key_path: str = os.path.join(kibana_certs_path, "kibana.key")
# setting up the secure connection using TLS
context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=server_cert_path)
context.load_cert_chain(certfile=client_cert_path, keyfile=client_key_path)
connections.create_connection(hosts=f"{scheme}://{user}:{pw}@{host}:{port}", ssl_context=context)


class Redner(Document):
    """ Every instance of this class represents a single Document in ElasticSearch, with various fields. """
    fraktion: Text = Text()
    nachname: Text = Text()
    ortszusatz: Text = Text()
    rolle_kurz: Text = Text()
    rolle_lang: Text = Text()
    titel: Text = Text()
    vorname: Text = Text()

    class Index:
        """ This is the index that will be created in ElasticSearch. We will use it later for querying the data. """
        name = speaker_index
        settings = {
            "number_of_shards": 2,
        }

    def save(self, **kwargs):
        """ Save a new Document in the ElasticSearch index. """
        return super(Redner, self).save(**kwargs)


class Dbtplenarprotokoll(Document):
    """ Every instance of this class represents a single Document in ElasticSearch, with various fields. """
    sitzung_datum: Date = Date()
    zitate_anzahl: Integer = Integer()

    class Index:
        """ This is the index that will be created in ElasticSearch. We will use it later for querying the data. """
        name = protocol_index
        settings = {
            "number_of_shards": 2,
        }

    def save(self, **kwargs):
        """ Save a new Document in the ElasticSearch index. """
        return super(Dbtplenarprotokoll, self).save(**kwargs)


def insert_citation_data() -> None:
    """ Adds new XML data to ElasticSearch based on a class definition, i.e. a manually defined schema. """
    # create the mappings in elasticsearch
    Dbtplenarprotokoll.init()
    print("Ingesting XML data from files...")
    for file_name in [x for x in os.listdir(protocols_folder_path) if x.endswith(".xml")]:
        file_path: str = os.path.join(protocols_folder_path, file_name)
        # get the root element from the XML file
        dbtplenarprotokoll: etree._Element = etree.fromstring(open(file_path, "rb").read())
        date_string: str = dbtplenarprotokoll.get("sitzung-datum")
        date_python: datetime = datetime.strptime(date_string, "%d.%m.%Y")
        protocol_id: str = dbtplenarprotokoll.get("sitzung-nr")
        citation_elements: List[etree._Element] = dbtplenarprotokoll.findall(".//p[@klasse='Z']")
        citation_count: int = len(citation_elements)
        new_protocol: Dbtplenarprotokoll = Dbtplenarprotokoll(meta=dict(id=protocol_id), sitzung_datum=date_python,
                                                              zitate_anzahl=citation_count)
        new_protocol.save()


def insert_speaker_data() -> None:
    """ Adds new XML data to ElasticSearch based on a class definition, i.e. a manually defined schema. """
    # create the mappings in elasticsearch
    Redner.init()
    # get the root element from the XML file
    dbtplenarprotokoll: etree._Element = etree.fromstring(open(f"{protocols_folder_path}/19084-data.xml", "rb").read())
    # find all speaker elements
    redner_xml_list: List[etree._Element] = dbtplenarprotokoll.findall(".//redner")
    for redner_xml in redner_xml_list:
        # get all available property names from our class
        props: List[str] = list(Redner.__annotations__)
        props_dict: Dict[str, Any] = dict()
        for prop in props:
            # property names are mapped 1-to-1 on the respective XML elements
            prop_xml: etree._Element = redner_xml.find(f".//{prop}")
            if prop_xml is not None:
                # fill the property dictionary with text values from the XML data
                props_dict[prop] = prop_xml.text
        # create and save a speaker
        new_redner: Redner = Redner(meta=dict(id=redner_xml.get("id")), **props_dict)
        new_redner.save()


def search_citation_data() -> None:
    """ Searches for relevant data in the ElasticSearch index. """
    # default sort order is "ascending", i.e. oldest dates are first in the list
    search: Search = Search(index=protocol_index).query("match_all").sort("sitzung_datum")
    # Python-style queries can be transformed to regular JSON format, which is usable in the Kibana Dev Tools etc.
    print("Query: ", search.to_dict())
    # .execute() only shows the first 10 results, so we use .scan() to find them all
    protocols: Iterator[Dbtplenarprotokoll] = search.scan()
    protocols_list: List[Dbtplenarprotokoll] = list(protocols)
    # searches with .scan() are not sorted by default, so we need to sort them explicitly
    protocols_list.sort(key=lambda x: x.sitzung_datum)
    print("Datum\t\t\tZitate")
    print("\n".join([f"{x.sitzung_datum}\t{x.zitate_anzahl}" for x in protocols_list]))


def search_speaker_data() -> None:
    """ Searches for relevant data in the ElasticSearch index. """
    search: Search = Search(index=speaker_index).query("match", titel="Dr.")
    # Python-style queries can be transformed to regular JSON format, which is usable in the Kibana Dev Tools etc.
    print("Query: ", search.to_dict())
    # .execute() only shows the first 10 results, so we use .scan() to find them all
    speakers: Iterator[Redner] = search.scan()
    counter: Counter = Counter()
    # count the number of hits for each party
    counter.update([x.fraktion for x in speakers])
    print("Which party has the most PhD-level speakers?", counter.most_common())


# feel free to comment out the "insert_..." methods after they have been executed once (successfully)
insert_speaker_data()
search_speaker_data()

insert_citation_data()
search_citation_data()
